module Database where

import Core

import Control.Exception          (IOException)
import Control.Monad.Catch        (catch)
import Data.Aeson                 (ToJSON)
import Data.ByteString.Lazy.UTF8  (fromString)
import Data.Maybe                 (listToMaybe)
import Database.PostgreSQL.Simple 
import Effectful
import Effectful.Dispatch.Dynamic
import Effectful.Error.Static     (Error, HasCallStack, catchError, runErrorNoCallStack, throwError)
import Servant hiding             ((:>), throwError)

runDatabaseDebug :: (IOE :> es, Error ServerError :> es) => Eff (Database : es) a -> Eff es a
runDatabaseDebug = interpret $ \_ -> \case
  DatabaseRead  (statement, values)  -> adapt $ read  statement values
  DatabaseWrite (statement, values)  -> adapt $ write statement values
  where
    read  _ values = 
      putStrLn "We just mocked a READ database operation" >>= \_ -> pure $ 
        Just (User values "Mock User")
    write _ values = 
      putStrLn $ "We just mocked a WRITE database operation with a user named " 
        ++ values

runDatabaseIO :: (IOE :> es, Error ServerError :> es) => Eff (Database : es) a -> Eff es a
runDatabaseIO = interpret $ \_ -> \case
  DatabaseRead  (statement, values) -> adapt $ read  statement values
  DatabaseWrite (statement, values) -> adapt $ write statement values
  where
    read  :: Query -> Int -> IO (Maybe User)
    read statement values  = do
      conn <- openConn
      user <- query conn statement (Only values)
      pure  $ listToMaybe user
    write :: Query -> String -> IO ()
    write statement values = do
      conn <- openConn
      execute conn statement (Only values)
      putStrLn $ "Wrote user to database using statement:\n" ++ show statement
    openConn = connectPostgreSQL "host=localhost port=5432 dbname=demo connect_timeout=10"

adapt :: (IOE :> es, Error ServerError :> es) => IO a -> Eff es a
adapt m = liftIO m `catch` \(e::IOException) ->
  throwError $ ServerError
    { errHTTPCode     = 500
    , errReasonPhrase = "Internal Database Error"
    , errBody         = fromString $ show e
    , errHeaders      = []
    }

queryUser :: Int -> (Query, Int)
queryUser userId = ("SELECT id, name FROM users WHERE id = ?;", userId)

writeUser :: String -> (Query, String)
writeUser name   = ("INSERT INTO users (name) VALUES (?);", name)
